/*
    Generated by sbus code generator

    Copyright (C) 2017 Red Hat

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "sbus/sbus_interface_declarations.h"
#include "sbus/interface_dbus/sbus_dbus_symbols.h"

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_AddMatch = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "rule"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_GetConnectionUnixProcessID = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "u", .name = "pid"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_GetConnectionUnixUser = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "u", .name = "uid"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_GetId = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "s", .name = "bus_name"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_GetNameOwner = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "s", .name = "unique_name"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Hello = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_ListActivatableNames = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "as", .name = "names"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_ListNames = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "as", .name = "names"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_ListQueuedOwners = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "as", .name = "unique_names"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_NameHasOwner = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "b", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_ReleaseName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "u", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_RemoveMatch = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "rule"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_RequestName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {.type = "u", .name = "flags"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "u", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_StartServiceByName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {.type = "u", .name = "flags"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "u", .name = "result"},
        {NULL}
    }
};

const struct sbus_argument
_sbus_dbus_args_org_freedesktop_DBus_NameAcquired[] = {
    {.type = "s", .name = "name"},
    {NULL}
};

const struct sbus_argument
_sbus_dbus_args_org_freedesktop_DBus_NameLost[] = {
    {.type = "s", .name = "name"},
    {NULL}
};

const struct sbus_argument
_sbus_dbus_args_org_freedesktop_DBus_NameOwnerChanged[] = {
    {.type = "s", .name = "name"},
    {.type = "s", .name = "new_owner"},
    {.type = "s", .name = "old_owner"},
    {NULL}
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Introspectable_Introspect = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "s", .name = "introspection"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Properties_Get = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "interface_name"},
        {.type = "s", .name = "property_name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "v", .name = "property_value"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Properties_GetAll = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "interface_name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "a{sv}", .name = "properties"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Properties_Set = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "interface_name"},
        {.type = "s", .name = "property_name"},
        {.type = "v", .name = "new_value"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {NULL}
    }
};
