/*
    SSSD

    Pack user info messages

    Authors:
        Sumit Bose <sbose@redhat.com>

    Copyright (C) 2009 Red Hat

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef __USER_INFO_MSG_H__
#define __USER_INFO_MSG_H__


errno_t pack_user_info_chpass_error(TALLOC_CTX *mem_ctx,
                                    const char *user_error_message,
                                    size_t *len,
                                    uint8_t **_resp);

#endif /* __USER_INFO_MSG_H__ */
