/*
    Generated by sbus code generator

    Copyright (C) 2017 Red Hat

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _SBUS_IFP_SYMBOLS_H_
#define _SBUS_IFP_SYMBOLS_H_

#include "sbus/sbus_interface_declarations.h"

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_FindBackendByName;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_FindDomainByName;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_FindMonitor;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_FindResponderByName;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_GetUserAttr;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_GetUserGroups;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_ListBackends;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_ListComponents;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_ListDomains;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_ListResponders;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Ping;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Cache_List;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Cache_ListByDomain;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Cache_Object_Remove;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Cache_Object_Store;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_ActiveServer;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_IsOnline;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_ListServers;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_ListServices;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_RefreshAccessRules;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_FindByID;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_FindByName;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_ListByDomainAndName;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_ListByName;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_Group_UpdateMemberList;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByCertificate;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByID;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByName;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByNameAndCertificate;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByValidCertificate;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_ListByAttr;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_ListByCertificate;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_ListByDomainAndName;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_ListByName;

extern const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_User_UpdateGroupsList;

#endif /* _SBUS_IFP_SYMBOLS_H_ */
