/*
    Generated by sbus code generator

    Copyright (C) 2017 Red Hat

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "sbus/sbus_interface_declarations.h"
#include "responder/ifp/ifp_iface/sbus_ifp_symbols.h"

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_FindBackendByName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "backend"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_FindDomainByName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "domain"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_FindMonitor = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "monitor"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_FindResponderByName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "responder"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_GetUserAttr = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "user"},
        {.type = "as", .name = "attr"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "a{sv}", .name = "values"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_GetUserGroups = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "user"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "as", .name = "values"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_ListBackends = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "backends"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_ListComponents = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "components"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_ListDomains = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "domain"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_ListResponders = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "responders"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Ping = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "ping"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "s", .name = "pong"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Cache_List = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Cache_ListByDomain = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "domain_name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Cache_Object_Remove = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "b", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Cache_Object_Store = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "b", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_ActiveServer = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "service"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "s", .name = "server"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_IsOnline = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "b", .name = "status"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_ListServers = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "service_name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "as", .name = "servers"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_ListServices = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "as", .name = "services"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Domains_Domain_RefreshAccessRules = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_FindByID = {
    .input = (const struct sbus_argument[]){
        {.type = "u", .name = "id"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_FindByName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_ListByDomainAndName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "domain_name"},
        {.type = "s", .name = "name_filter"},
        {.type = "u", .name = "limit"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_ListByName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name_filter"},
        {.type = "u", .name = "limit"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Groups_Group_UpdateMemberList = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByCertificate = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "pem_cert"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByID = {
    .input = (const struct sbus_argument[]){
        {.type = "u", .name = "id"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByNameAndCertificate = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name"},
        {.type = "s", .name = "pem_cert"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_FindByValidCertificate = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "pem_cert"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "o", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_ListByAttr = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "attribute"},
        {.type = "s", .name = "attr_filter"},
        {.type = "u", .name = "limit"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_ListByCertificate = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "pem_cert"},
        {.type = "u", .name = "limit"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_ListByDomainAndName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "domain_name"},
        {.type = "s", .name = "name_filter"},
        {.type = "u", .name = "limit"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_ListByName = {
    .input = (const struct sbus_argument[]){
        {.type = "s", .name = "name_filter"},
        {.type = "u", .name = "limit"},
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {.type = "ao", .name = "result"},
        {NULL}
    }
};

const struct sbus_method_arguments
_sbus_ifp_args_org_freedesktop_sssd_infopipe_Users_User_UpdateGroupsList = {
    .input = (const struct sbus_argument[]){
        {NULL}
    },
    .output = (const struct sbus_argument[]){
        {NULL}
    }
};
