/*
    Generated by sbus code generator

    Copyright (C) 2017 Red Hat

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _SBUS_DBUS_SYMBOLS_H_
#define _SBUS_DBUS_SYMBOLS_H_

#include "sbus/sbus_interface_declarations.h"

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_AddMatch;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_GetConnectionUnixProcessID;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_GetConnectionUnixUser;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_GetId;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_GetNameOwner;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Hello;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_ListActivatableNames;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_ListNames;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_ListQueuedOwners;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_NameHasOwner;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_ReleaseName;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_RemoveMatch;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_RequestName;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_StartServiceByName;

extern const struct sbus_argument
_sbus_dbus_args_org_freedesktop_DBus_NameAcquired[];

extern const struct sbus_argument
_sbus_dbus_args_org_freedesktop_DBus_NameLost[];

extern const struct sbus_argument
_sbus_dbus_args_org_freedesktop_DBus_NameOwnerChanged[];

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Introspectable_Introspect;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Properties_Get;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Properties_GetAll;

extern const struct sbus_method_arguments
_sbus_dbus_args_org_freedesktop_DBus_Properties_Set;

#endif /* _SBUS_DBUS_SYMBOLS_H_ */
